#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "dSFMT.h"

///////////////////////////////////////////////////////////////////////////////
// TODO: Do not use global variables and separate group of functions 
//       into different files
///////////////////////////////////////////////////////////////////////////////

#define CONFIG_FILE "config.txt"
#define GNUPLOT_FILE "generateGraphs.plot"
#define STRING_LIMIT 256
#define TRUE 1
#define FALSE 0


typedef struct {
    char output_name[STRING_LIMIT];
    int montecarlo_loops;
    int sample_num;

    // Lattice size, adjust
    int nx;
    int ny;

    double temp_min;
    double temp_max;
    int temp_measurements; // Number of measurement in the interval above
    int constant_magnetic_coupling; // TRUE or FALSE
    double magnetic_coupling;
    int constant_magnetic_field; // TRUE or FALSE
    double magnetic_field;
} configuration;

///////////////////////////////////////////////////////////////////////////////
/// BEGIN: Global Variables
configuration config;

// Spin, magnetic coupling and magnetic field at site x,y
int **s;
double **J;
double **h;

double **vEnergySample;
double **vEnergySquareSample;
double **vMagnetisationSample;

double *vMagnetisation;
double *vTemperature;
double *vMeanEnergyEstimator;
double *vEnergyEstimator; // Needed to calculate Specific Heat
double *vSpecificHeat;
/// END: Global Variables
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
/// BEGIN: Matrix manipulation functions
int xUp(int i) {
    return ((i+1) % config.nx);
}

int xDown(int i) {
    return ((i-1+config.nx) % config.nx);
}
int yUp(int i) {
    return ((i+1) % config.ny);
}
int yDown(int i) {
    return ((i-1+config.ny) % config.ny);
}


int** create_imatrix(int size_x, int size_y) {
    int i;
    int** matrix;
    matrix = calloc(size_x, sizeof(int*));
    for(i = 0;i<size_x;i++) {
        matrix[i] = calloc(size_y, sizeof(int));
    }
    return matrix;
}

double** create_dmatrix(int size_x, int size_y) {
    int i;
    double** matrix;
    matrix = calloc(size_x, sizeof(double*));
    for(i = 0;i<size_x;i++) {
        matrix[i] = calloc(size_y, sizeof(double));
    }
    return matrix;
}

void destroy_matrix(void** matrix) {
    int i;
    for( i=0; matrix[i] != NULL; i++ ) {
        free( matrix[i] );
    }
    free( matrix );
}
/// END: Matrix manipulation functions
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
/// BEGIN: Monte carlo functions

// Create data structures and initialize spin state (s), magnetic coupling (J) and magnetic field (h)
void initialize(dsfmt_t *dsfmt){
    int x, y;

    // Structures creation
    s = create_imatrix(config.nx, config.ny);
    J = create_dmatrix(config.nx, config.ny);
    h = create_dmatrix(config.nx, config.ny);

    vEnergySample = create_dmatrix(config.temp_measurements, config.sample_num);
    vEnergySquareSample = create_dmatrix(config.temp_measurements, config.sample_num);
    vMagnetisationSample = create_dmatrix(config.temp_measurements, config.sample_num);

    vMagnetisation = malloc(config.temp_measurements*sizeof(double));
    vTemperature = malloc(config.temp_measurements*sizeof(double));
    vMeanEnergyEstimator = malloc(config.temp_measurements*sizeof(double));
    vEnergyEstimator = malloc(config.temp_measurements*sizeof(double));
    vSpecificHeat = malloc(config.temp_measurements*sizeof(double));


    for (x=0; x<config.nx; x++) {
        for (y=0; y<config.ny; y++) {
            // Generate a random number distributed uniformly between [0,1) for selecting spin status
            // Initial state: highly disordered (hot)
            if (dsfmt_genrand_close_open(dsfmt) < 0.5) {
                s[x][y] = 1;
            } else {
                s[x][y] = -1;
            }

            if (config.constant_magnetic_coupling == TRUE) {
                // Constant magnetic coupling for every bond
                J[x][y] = config.magnetic_coupling; 
            } else {
                // Generate a random X coordinate between [0,5)
                J[x][y] = dsfmt_genrand_close_open(dsfmt)*5;
            }
            if (config.constant_magnetic_field  == TRUE) {
                // Constant magnetic field for every field
                h[x][y] = config.magnetic_field;
            } else {
                // Generate a random X coordinate between [0,5)
                h[x][y] = dsfmt_genrand_close_open(dsfmt)*5; 
            }
        }
    }

}

// Modify current state with a given probability that depends on the amount of energy change and temperature
void update_state(dsfmt_t *dsfmt, double temp) {
    int x,y;
    double p_plus,p_minus;
    int current_spin, new_spin;
    double H, new_H, H_change;
    double random_var;
    
    x = (int)(dsfmt_genrand_close_open(dsfmt) * 10) % config.nx;
    y = (int)(dsfmt_genrand_close_open(dsfmt) * 10) % config.ny;
    
    current_spin = s[x][y];
    new_spin = - s[x][y];
    
    // Current energy at chosen point
    double a = J[xUp(x)][y] * current_spin * s[xUp(x)][y];
    double b = J[xDown(x)][y] * current_spin * s[xDown(x)][y];
    double c = J[x][yUp(y)] * current_spin * s[x][yUp(y)];
    double d = J[x][yDown(y)] * current_spin * s[x][yDown(y)];
    // NOTE: Double counting is not produced because we are measuring a single point,
    //       not the whole matrix
    if (config.nx == 1 && config.ny == 1) {
        H = - h[x][y]*current_spin;
    } else if (config.nx == 1){
        H = - (c + d) - h[x][y]*current_spin; 
    } else if (config.ny == 1){
        H = - (a + b) - h[x][y]*current_spin;
    } else {
        H = - (a + b + c + d) - h[x][y]*current_spin;
    }

    // New energy at chosen point
    a = J[xUp(x)][y] * new_spin * s[xUp(x)][y];
    b = J[xDown(x)][y] * new_spin * s[xDown(x)][y];
    c = J[x][yUp(y)] * new_spin * s[x][yUp(y)];
    d = J[x][yDown(y)] * new_spin * s[x][yDown(y)];
    // NOTE: Double counting is not produced because we are measuring a single point,
    //       not the whole matrix
    if (config.nx == 1 && config.ny == 1) {
        new_H = - h[x][y]*new_spin;
    } else if (config.nx == 1){
        new_H = - (c + d) - h[x][y]*new_spin;
    } else if (config.ny == 1){
        new_H = - (a + b) - h[x][y]*new_spin;
    } else {
        new_H = - (a + b + c + d) - h[x][y]*new_spin;
    }

    // Contribution to total energy if change is made
    H_change = new_H - H;
    
    /*if (H_change < 0) {
        // New state is less energetic
        s[x][y] = new_spin;
    } else {*/
        random_var = dsfmt_genrand_close_open(dsfmt); // random number 0-1
        if (random_var < exp(-H_change/temp)) {
            s[x][y] = new_spin;
        }
    //}
}
/// END: Monte carlo functions
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
/// BEGIN: File functions


void create_gnuplotfile(){
    FILE *file;

    chdir("output");
    file = fopen(GNUPLOT_FILE,"w");
    chdir("..");
    fprintf(file, "#!/usr/bin/gnuplot\n");
    fprintf(file, "set grid\n");
    fprintf(file, "set xlabel 'Temperature'\n");
    fprintf(file, "unset key\n");
    fprintf(file, "set rmargin 10\n");
    fprintf(file, "set lmargin 10\n");
    fprintf(file, "set terminal png size 800, 600\n");
    fprintf(file, "\n");
    
    fclose(file);
}

void append_to_gnuplotfile(char *output_name, int xmin, int xmax){
    FILE *file;

    chdir("output");
    file = fopen(GNUPLOT_FILE,"a");
    chdir("..");
    fprintf(file, "set xrange [%i:%i]\n", xmin, xmax);
    fprintf(file, "set xtics %f\n", (xmax - xmin) / 10.0);
    fprintf(file, "set title 'Mean Energy (%s)'\n", output_name);
    fprintf(file, "set ylabel '<H>'\n");
    fprintf(file, "set output \"%s-MeanEnergy.png\"\n", output_name);
    fprintf(file, "plot \"%s.txt\" every ::1 using 2:3 with linespoints pt 6 title 'Mean energy'\n", output_name);
    fprintf(file, "\n");
    fprintf(file, "set title 'Specific Heat (%s)'\n", output_name);
    fprintf(file, "set ylabel 'Cv'\n");
    fprintf(file, "set output \"%s-SpecificHeat.png\"\n", output_name);
    fprintf(file, "plot \"%s.txt\" every ::2 using 2:4 with linespoints pt 6 title 'Specific Heat'\n", output_name);
    fprintf(file, "\n");
    fprintf(file, "set title 'Magnetisation (%s)'\n", output_name);
    fprintf(file, "set ylabel 'M'\n");
    fprintf(file, "set output \"%s-Magnetisation.png\"\n", output_name);
    fprintf(file, "plot \"%s.txt\" every ::1 using 2:5 with linespoints pt 6 title 'Magnetisation'\n", output_name);
    fprintf(file, "\n\n\n");
    fclose(file);
}

// Read config file with one or more config sets and extract one
configuration *read_config(FILE *file) {
    configuration config = {"default", 1000, 10, 2 , 2, 0.1, 5, 10, FALSE, 0, FALSE, 0};
    configuration *pconfig = &config;
    int level;
    char line[STRING_LIMIT], *param, *value;
    fpos_t pos;

    config.constant_magnetic_coupling = FALSE;
    config.constant_magnetic_field = FALSE;

    while (!feof(file)) {
        fgetpos(file, &pos); // Position before reading

        if (fgets(line, sizeof(line), file) == NULL) {
            continue;
        }
        // Ignore comments and empty lines
        if (line[0] != '#' && strlen(line) > 1) {
            param = strtok(line, "=");
            value = strtok(NULL, "\n"); // Remove \n from line
            
            if (strcmp(param, "OUTPUT_NAME") == 0) {

                if ((strcmp(config.output_name, "default") == 0)) { // First time: config begins
                    strcpy(config.output_name, value);
                } else {    // New config: end reading
                    fsetpos(file, &pos); // Go back to last position
                    break;
                }
                
            } else if (strcmp(param, "MONTECARLO_LOOPS") == 0) {
                config.montecarlo_loops = atoi(value);
            } else if (strcmp(param, "SAMPLE_NUM") == 0) {
                config.sample_num = atoi(value);
            } else if (strcmp(param, "NX") == 0) {
                config.nx = atoi(value);
            } else if (strcmp(param, "NY") == 0) {
                config.ny = atoi(value);
            } else if (strcmp(param, "TEMP_MIN") == 0) {
                config.temp_min = atof(value);
            } else if (strcmp(param, "TEMP_MAX") == 0) {
                config.temp_max = atof(value);
            } else if (strcmp(param, "TEMP_MEASUREMENTS") == 0) {
                config.temp_measurements = atoi(value);
            } else if (strcmp(param, "CONSTANT_MAGNETIC_COUPLING") == 0) {
                config.constant_magnetic_coupling = TRUE;
                config.magnetic_coupling = atof(value);
            } else if (strcmp(param, "CONSTANT_MAGNETIC_FIELD") == 0) {
                config.constant_magnetic_field = TRUE;
                config.magnetic_field = atof(value);
            }
        }
    }

    if ((strcmp(config.output_name, "default") == 0)) {
        // No more configurations found in file
        return NULL;
    } else {
        return pconfig;
    }
}

// Save into matrices the measurements needed for a given sample
void register_sample_measurement(int temperature_level, int sample, double temperature) {
    // Calculate mean energy and magnetization, and print out
    int x,y;
    double magnetisation,H, Ha, Hb;
    
    // Sum over the neighbour sites
    H = magnetisation = Ha = Hb = 0.0;

    for (x=0; x<config.nx; x++) {
        for (y=0; y<config.ny; y++) {
          
            // Calculate magnetic coupling
            double a = J[xUp(x)][y] * s[x][y] * s[xUp(x)][y];
            double b = J[xDown(x)][y] * s[x][y] * s[xDown(x)][y];
            double c = J[x][yUp(y)] * s[x][y] * s[x][yUp(y)];
            double d = J[x][yDown(y)] * s[x][y] * s[x][yDown(y)];
            // NOTE: Double counting is produced because we pass through the
            //       whole matrix and every link will be sum twice
            if (config.nx == 1 && config.ny == 1) {
                Ha += 0;
            } else if (config.nx == 1){
                Ha += (c + d)/2; // Correct double counting dividing by 2
            } else if (config.ny == 1){
                Ha += (a + b)/2; // Correct double counting dividing by 2
            } else {
                Ha += (a + b + c + d)/2; // Correct double counting dividing by 2
            }
            
            Hb += h[x][y]*s[x][y];
            magnetisation += s[x][y];
        }
    }

    H = - Ha - Hb;  // Energy
    //H = H/(config.nx*config.ny);  // Mean energy
    magnetisation = magnetisation/(config.nx*config.ny);

    vEnergySample[temperature_level][sample] = H;
    vEnergySquareSample[temperature_level][sample] = H*H;
    vMagnetisationSample[temperature_level][sample] = magnetisation;
}

// Sum up measurements for a group of sample and save the results into arrays 
void register_measurement(int temperature_level, double temperature) {
    // Calculate mean energy and magnetization, and print out
    int sample;
    double H, H2, magnetization;

    /* sum over the neighbour sites - typewriter fashion */
    H = H2 = magnetization = 0.0;
    for (sample=0; sample<config.sample_num; sample++) {
        H += vEnergySample[temperature_level][sample];
        H2 += vEnergySquareSample[temperature_level][sample];
        magnetization += vMagnetisationSample[temperature_level][sample];
    }

    vTemperature[temperature_level] = temperature;

    vMagnetisation[temperature_level] = fabs(magnetization/(config.sample_num));
    vMeanEnergyEstimator[temperature_level] = H/(config.sample_num);
    vSpecificHeat[temperature_level] = ( (H2/(config.sample_num)) - pow(H/(config.sample_num),2)) / pow(temperature,2);    
}

// Print results and save them into an output file
void save_measurements(){
    int level;
    FILE *file; 

    append_to_gnuplotfile(config.output_name, config.temp_min, config.temp_max);

    chdir("output");
    file = fopen(strcat(config.output_name, ".txt"),"w");
    chdir("..");
    printf("* Next output: %s\n", config.output_name);
    printf("---------------------------------------------------------------------\n");
    printf("Level\tTemperature\tMeanEnergyEst\tSpecificHeat\tMagnetization\n");
    printf("---------------------------------------------------------------------\n");

    fprintf(file, "Level\tTemperature\tMeanEnergyEst\tSpecificHeat\tMagnetization\n");
    for (level=0; level<config.temp_measurements; level++) {
        fprintf(file, "%i\t%f\t%f\t%f\t%f\n", level, vTemperature[level], vMeanEnergyEstimator[level], vSpecificHeat[level], vMagnetisation[level]);
        printf("%i\t%f\t%f\t%f\t%f\n", level, vTemperature[level], vMeanEnergyEstimator[level], vSpecificHeat[level], vMagnetisation[level]);
    }
    printf("---------------------------------------------------------------------\n");
    fclose(file);
}
/// END: File functions
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
/// BEGIN: MAIN
int main(int argc, char* argv[]) {
    FILE *file;
    dsfmt_t dsfmt; // Random number generator
    int i, sample, temperature_level;
    double temperature;
    int seed = 12345;

    create_gnuplotfile();

    // For each config set in the config file
    chdir("input");
    file = fopen(CONFIG_FILE,"r");
    chdir("..");
    configuration *pconfig = read_config(file);
    while (pconfig != NULL) {
        memcpy(&config, pconfig, sizeof(config)); // Make read config global

        // Initialize the random number generator (RNG)
        dsfmt_init_gen_rand(&dsfmt, seed);

        temperature = config.temp_min;
        if (temperature == 0) temperature += 0.00001; // Avoid zero division
        temperature_level = 0;

        // Create data structures and initialize spin state (s), magnetic coupling (J) and magnetic field (h)
        initialize(&dsfmt);

        // For each temperature
        while (temperature < config.temp_max && temperature_level < config.temp_measurements) {
        
            // Generate samples
            for (sample=0; sample < config.sample_num; sample++) {
                // Monte carlo simulation for a given sample
                for (i=0; i < config.montecarlo_loops; i++) {
                    update_state(&dsfmt, temperature);
                }
                register_sample_measurement(temperature_level, sample, temperature);
            }

            // Calculate mean values for the group of samples at a given temperature
            register_measurement(temperature_level, temperature);
            
            temperature_level++;
            temperature += ((double)config.temp_max - config.temp_min) / config.temp_measurements;
            if (temperature == 0) temperature += 0.00001; // Avoid zero division
        }

        // Print and save outputs
        save_measurements();

        // Next config set
        pconfig = read_config(file);
    }

    fclose(file);
    return 0;
}
/// END: MAIN
///////////////////////////////////////////////////////////////////////////////

